import torch

from TakeBlipNer.predict import NerPredict
from TakeBlipPosTagger.predict import PosTaggerPredict
from TakeBlipNer.utils import load_fasttext_embeddings


def load_postagging_predictor(postagging_model_path: str,
                              postagging_label_path: str,
                              embedding_path: str) -> PosTaggerPredict:
    """
    Parameters
    ----------
    postagging_model_path : str
            path to the postagging model
    postagging_label_path : str
            path to the labels dictionary of postag model
    embedding_path : str
            path to embedding model

    Returns
    -------
    PosTaggerPredict
        Return object to predict the postag labels
    """
    embedding = load_fasttext_embeddings(embedding_path, '<pad>')
    postag_model = torch.load(postagging_model_path)
    postag_predictor = PosTaggerPredict(
        model=postag_model,
        label_path=postagging_label_path,
        embedding=embedding
    )
    return postag_predictor


def load_ner_predictor(ner_model_path: str, ner_label_path: str,
                       postag_predictor: PosTaggerPredict) -> NerPredict:
    """
    Parameters
    ---------
    ner_model_path : str
            path to the ner model
    ner_label_path : str
            path to the labels dictionary of ner model
    postag_predictor : PosTaggerPredict
            Object to predict the postag labels

    Returns
    -------
    NerPredict
        Return object to predict the ner labels
    """
    ner_model = torch.load(ner_model_path)
    ner_predictor = NerPredict(
        model=ner_model,
        label_path=ner_label_path,
        postag_model=postag_predictor,
        pad_string='<pad>',
        unk_string='<unk>'
    )
    return ner_predictor



import time
import logging


class FileValidator:
    """Analyze whether the file is valid.
    
     Methods:
      * is_valid_file - Return `True` if file passes all validations, `False` otherwise.
      * has_valid_encoding - Return `True` if file has valid encoding, `False` otherwise.
      * has_valid_extension - Return `True` if file has valid extension, `False` otherwise.
      * has_empty_lines - Return `True` if file has empty lines, `False` otherwise.
      * get_error_message - Return message with found file errors.
    """

    def __init__(
        self, 
        encoding: str, 
        extension: str, 
        file_content: bytes, 
        file_name: str = 'file_name', 
        logger=None) -> None:
        """ 
        :param encoding: Correct file encoding.
        :type encoding: `str`
        :param extension: Correct file extension.
        :type extension: `str`
        :param file_content: File content.
        :type file_content: `bytes`
        :param file_name: File name.
        :type file_name: `str`
        :param logger: Object to log class activities.
        :type logger: `None`
        """
        self.__encoding = encoding
        self.__extension = extension
        self.__error_message = []
        self.__logger = logger
        self.__file_content = file_content
        self.__file_name = file_name

    def is_valid_file(self) -> bool:
        """Return `True` if file passes all validations, `False` otherwise.

        :return: `True` if file passes all validations, `False` otherwise.
        :rtype: `bool`
        """
        start_time = time.time()
        is_valid_file = all([
            self.has_valid_encoding(),
            self.has_valid_extension(),
            not self.has_empty_lines()
        ])
        self.__log('Time elapsed to validate file {} seconds'.format(time.time() - start_time))
        return is_valid_file

    def has_valid_encoding(self) -> bool:
        """Return `True` if file has valid encoding, `False` otherwise.

        :return: `True` if file has valid encoding, `False` otherwise.
        :rtype: `bool`
        """
        try:
            decoded_data = self.__file_content.strip().decode(self.__encoding)
            return True
        except UnicodeDecodeError as e:
            self.__error_message.append('Encoding inválido')
            self.__log_error(self.__error_message)
            return False

    def has_valid_extension(self) -> bool:
        """Return `True` if file has valid extension, `False` otherwise.

        :return: `True` if file has valid extension, `False` otherwise.
        :rtype: `bool`
        """
        file_type = self.__file_name.split('.')[-1]
        if self.__extension not in file_type:
            self.__error_message.append('Extensão inválida')
            self.__log_error(self.__error_message)
            return False
        return True

    def has_empty_lines(self) -> bool:
        """Return `True` if file has empty lines, `False` otherwise.
        
        :return: `True` if file has empty lines, `False` otherwise.
        :rtype: `bool`
        """
        try:
            empty_lines = sum(len(line.strip()) == 0 for line in self.__file_content.splitlines())
            if empty_lines == 0:
                return False
            self.__error_message.append('Linha em branco')
            self.__log_error(self.__error_message)
            return True
        except UnicodeDecodeError as e:
            self.__log_error(e)
            return True

    def get_error_message(self) -> str:
        """Return message with found file errors. 
        
        :return: message with found file errors.
        :rtype: `str`
        """
        found_errors = ' ; '.join(self.__error_message)
        return f'Sua requisição não será processada. Arquivo enviado possui o(s) seguinte(s) erro(s): {found_errors}'

    def __log_error(self, message: str) -> None:
        if self.__logger:
            self.__logger.log_error_message('ERROR', message)

    def __log(self, message: str) -> None:
        if self.__logger:
            self.__logger.log_message('DEBUG', message)



import os
import sys
import json
import uuid
import shutil
import logging
import traceback
from TakeBlipInsightExtractor.insight_extractor import InsightExtractor
from TakeBlipInsightExtractor.outputs.eventhub_log_sender \
    import EventHubLogSender


def read_json_file(root: str, file_name: str) -> dict:
    """Reads a .json file from root/file_name.

    Args:
       root (str): a string with file root path
       file_name (str): a string with file name

    Returns (Dict):
       A dictionary read .from json file
    """
    filename = os.path.join(root, file_name)
    try:
        with open(filename, mode='r') as f:
            return json.loads(f.read())
    except FileNotFoundError:
        return {}


secrets = read_json_file(os.path.abspath(os.path.dirname(__file__)),
                         'secrets.json')

class InsightExtractorExecutor:
    def __init__(self,
                 user_email,
                 bot_name,
                 embedding_model_path,
                 postagging_model_path,
                 postagging_label_path,
                 ner_model_path,
                 ner_label_path,
                 output_files_path):
        self.logger = None
        self.insight_extractor = None
        self.user_email = user_email
        self.bot_name = bot_name
        self.embedding_model_path = embedding_model_path
        self.postagging_model_path = postagging_model_path
        self.postagging_label_path = postagging_label_path
        self.ner_model_path = ner_model_path
        self.ner_label_path = ner_label_path
        self.output_files_path = output_files_path
        self.root_path = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                      'default_models')
        self.initialize_eventhub_logger()
        self.initialize_insight_extractor()

    def initialize_eventhub_logger(self):
        try:
            correlation_id = str(uuid.uuid3(uuid.NAMESPACE_DNS,
                                            self.user_email + self.bot_name))

            self.logger = EventHubLogSender(
                application_name='Insight Extractor Package',
                user_email=self.user_email,
                bot_name=self.bot_name,
                file_name='file_name',
                correlation_id=correlation_id,
                connection_string=secrets.get('EVENTHUB_CONNECTION_STRING'),
                eventhub_name=secrets.get('EVENTHUB_NAME')
            )
        except (Exception) as e:
            logging.error(
                'Error {} while initializing Event Hub Logger!'.format(e))

    def initialize_insight_extractor(self):
        self.insight_extractor = InsightExtractor(
            input_data='file',
            separator=',',
            similarity_threshold=0.65,
            embedding_path=self.embedding_model_path,
            postagging_model_path=self.postagging_model_path,
            postagging_label_path=self.postagging_label_path,
            ner_model_path=self.ner_model_path,
            ner_label_path=self.ner_label_path,
            user_email=self.user_email,
            bot_name=self.bot_name,
            use_neighborhood=True,
            logger=self.logger,
            datalake_connector=None,
        )

    def copy_files(self):
        for file in self.insight_extractor.output_files_dict.values():
            shutil.copyfile(os.path.join(os.getcwd(), file),
                            os.path.join(self.output_files_path, file))

    def remove_files(self):
        try:
            self.logger.log_message('DEBUG', 'Removing local files. ')
            for file in self.insight_extractor.output_files_dict.values():
                os.remove(file)
            self.logger.log_message('DEBUG', 'Local files removed. ')
        except Exception as e:
            self.logger.log_error_message('ERROR',
                                     'Error {} while removing files'.format(e))

    def analyze(self, messages):
        self.insight_extractor.predict(percentage_threshold=0.5,
                                  node_messages_examples=10,
                                  batch_size=100,
                                  messages=messages)
        self.copy_files()
        self.remove_files()






def get_response_message(file_name, user_email):
    response_message = (
        'Recebemos a sua requisição para o arquivo {} e dentro '
        'de alguns minutos enviaremos para o e-mail {} passado.'
    ).format(file_name, user_email)
    return response_message


import time
import json
import base64
import logging
import typing as tp
from databricks import sql


class DatabricksConnector:
    """
    This class establishes a connection to the data lake through a sql
    connector to databricks.

    Attributes:
    :message_column (str) - a string with the name of the column
    containing the messages
    :storage_date_column (str) - a string with the name of the column
    containing the time when messages were stored
    :identity_column (str) - a string with the name of the column
    containing bot identity
    :data_table (str) - a string with the data table name
    :start_date (str) - a string with the start date
    :end_date (str) - a string with the end date
    :bot_name (str) - a string with bot identity
    :is_subject_extraction (bool) -  a boolean value indicating if data is to subject extraction
    :subject_data_table (str) -  a string with subject data table
    :subject_bot_name_column (str) -  a string with subject bot name column
    :subject_message_id_column (str) -  a string with subject message id column
    :subject_bot_name (str) -  a string with subject bot name

    Methods:
    :consume_data - Return a list of consumed/received data from data lake.
    """

    def __init__(self,
                 server_hostname: str,
                 http_path: str,
                 access_token: str,
                 message_column: str,
                 storage_date_column: str,
                 identity_column: str,
                 data_table: str,
                 start_date: str,
                 end_date: str,
                 bot_name: str,
                 separator: str = '|',
                 is_subject_extraction: bool = False,
                 message_id_column: str = None,
                 subject_data_table: str = None,
                 subject_bot_name_column: str = None,
                 subject_message_id_column: str = None,
                 ) -> None:
        """Initializes the class setting the query to execute in data lake.

        :param server_hostname: string with the host name.
        :type server_hostname: `str`
        :param http_path: string with the path to receive http request.
        :type http_path: `str`
        :param access_token: string with Databricks access token.
        :type access_token: `str`
        :param separator: string with separator
        :type separator: `str`
        :param message_column: string with the name of the column
         containing the messages.
        :type message_column: `str`
        :param storage_date_column: string with the name of the column
        containing the time when messages were stored.
        :type storage_date_column: `str`
        :param data_table: string with the data table name.
        :type data_table: `str`
        :param start_date: string with the start date.
        :type start_date: `str`
        :param end_date: string with the end date.
        :type end_date: `str`
        :param bot_name: string with bot identity.
        :type bot_name: `str`
        :param is_subject_extraction: boolean value indicating if data is to subject extraction
        :type is_subject_extraction: `bool`
        :param message_id_column: string with the message id column.
        :type message_id_column: `str`
        :param subject_data_table: string with the subject data table name.
        :type subject_data_table: `str`
        :param subject_bot_name_column: string with the subject bot name column.
        :type subject_bot_name_column: `str`
        :param subject_message_id_column: string with the subject message id column.
        :type subject_message_id_column: `str`
        """
        self.__server_hostname = server_hostname
        self.__http_path = http_path
        self.__access_token = access_token
        self.separator = separator
        self.message_column = message_column
        self.storage_date_column = storage_date_column
        self.identity_column = identity_column
        self.data_table = data_table
        self.start_date = start_date
        self.end_date = end_date
        self.bot_name = bot_name
        self.message_id_column = message_id_column
        self.subject_data_table = subject_data_table
        self.subject_bot_name_column = subject_bot_name_column
        self.subject_message_id_column = subject_message_id_column

        if is_subject_extraction:
            self.__create_subject_extraction_query()
        else:
            self.__create_default_query()

        print(self.query)

    def __create_subject_extraction_query(self) -> None:
        """Create a query string to be executed for subject extraction
        """
        try:
            self.query = f"""SELECT DISTINCT ChatBot, Content
            FROM (SELECT {self.subject_bot_name_column} AS ChatBot, {self.subject_message_id_column} 
                FROM {self.subject_data_table} 
                WHERE 1=1 AND
                ChatBot='{self.bot_name}' AND
                ToDomain='blip.ai' AND
                TicketId is not null
                    AND {self.storage_date_column} >= '{self.start_date}'
                    AND {self.storage_date_column} < '{self.end_date}') AS human_attendance
            LEFT JOIN
            (SELECT {self.message_column} AS Content, {self.message_id_column}
                FROM {self.data_table}
                WHERE 1=1 
                    AND {self.storage_date_column} >= '{self.start_date}'
                    AND {self.storage_date_column} < '{self.end_date}') AS messages
            ON human_attendance.{self.subject_message_id_column} = messages.{self.message_id_column}
            ORDER BY Content"""
        except Exception as e:
            print(e)

    def __create_default_query(self) -> None:
        """Create default query string to be executed
        """
        try:
            self.query = f"""
                SELECT {self.message_column} FROM {self.data_table}
                WHERE 1=1
                  AND {self.storage_date_column} >= '{self.start_date}'
                  AND {self.storage_date_column} < '{self.end_date}'
                  AND {self.identity_column} = '{self.bot_name}'
                """
        except Exception as e:
            print(e)

    def consume_data(self) -> tp.List[str]:
        """Consume data from a executed query in data lake

        :return: A list of collected data from datalake
        :rtype: `list`
        """

        if not self.query:
            raise Exception('Query not initialized')

        start_time = time.time()

        with sql.connect(server_hostname=self.__server_hostname,
                         http_path=self.__http_path,
                         access_token=self.__access_token) as connection:
            logging.debug('Data lake connection established.')
            with connection.cursor() as cursor:
                logging.debug('Executing query.')

                try:
                    cursor.execute(self.query)
                except Exception as e:
                    print(e)
                print('query executed')
                logging.debug('Query executed.')
                logging.debug('Fetching result.')
                result = cursor.fetchall()
                logging.debug('Result data received.')
                print('data received')
        logging.debug(f'Time elapsed in data consuming '
                      f'{time.time() - start_time}')
        print('messages colected from datalake')
        return self.extract_messages(result)

    def convertFromBase64(self, text) -> str:
        """Converts a text string from base64

        Parameters
        ----------
        text : str
            text string to convert from base64

        Returns
        -------
        str
            converted text
        """
        try:
            converted_texto = str(base64.b64decode(str(text))
                                  .decode('latin-1').replace('\x00', ''))
        except Exception as e:
            return 'Deleted'
        return converted_texto

    def convertFromUTF16(self, text: str) -> str:
        """Converts a text string from UTF-16

        Parameters
        ----------
        text : str
            text string to convert from UTF-16

        Returns
        -------
        str
            converted text
        """
        try:
            converted_text = str(base64.b64decode(str(text))
                        .decode('utf-16'))
        except Exception as e:
            return 'Deleted'
        return converted_text

    def isValidJson(self, text: str) -> int:
        """Validates if text message is a json

         Parameters
        ----------
        text : str
            Text string to convert

        Returns
        -------
        int:
            Integer value meaning if the message is a json
        """
        if text == 'Deleted':
            return 1
        try:
            _ = json.loads(text)
        except Exception as e:
            return 0
        return 1

    def extract_messages(self, result: tp.List[dict]) -> tp.List[tp.List[str]]:
        """Extract messages from datalake receveid data

        :return: A list of collected messages from received data
        :rtype: `list`
        """
        return [[self.convertFromUTF16(row[self.message_column]).replace(self.separator, '')]
                for row in result if len(row[self.message_column].strip()) > 0 and
                self.isValidJson(self.convertFromUTF16(row[self.message_column])) != 1]



import numpy as np
import matplotlib.colors as colors

from colorsys import hls_to_rgb
from wordcloud import WordCloud


def create_word_cloud(entity_frequency_dict: dict,
                      parent_entity_dict: dict,
                      background_color: str = 'white',
                      width: int = 1000,
                      height: int = 1000,
                      max_words = 200):
    """

    Create the word cloud for the entities, with the size of the word
    been the frequency of the entity and the color of the word based on
    the cluster.

    :param entity_frequency_dict: dictionary with the relation entity-frequency
    :type entity_frequency_dict: ``dict``
    :param parent_entity_dict: dictionary with the relation parent-entities
    :type parent_entity_dict: ``dict``
    :param background_color: background color of the word cloud
    :type background_color: ``str``
    :param width: the width of the word cloud
    :type width: ``int``
    :param height: the height of the word cloud
    :type height: ``int``
    :return: the word cloud object
    """
    show_words_dict = dict(sorted(entity_frequency_dict.items(),
                                  key=lambda item: item[1],
                                  reverse=True)[:max_words])

    wc = WordCloud(background_color=background_color,
                   width=width,
                   height=height).generate_from_frequencies(
        show_words_dict)

    filtered_parent_entity_dict = filter_parents(show_words_dict,
                                                 parent_entity_dict)
    cluster_color_dict = create_dict_colors(filtered_parent_entity_dict)
    grouped_color_func = SimpleGroupedColorFunc(cluster_color_dict)
    wc.recolor(color_func=grouped_color_func)
    return wc


def filter_parents(dict_words, dict_parent):
    filtered_dict = {}
    for parent, entities in dict_parent.items():
        if any([entity in dict_words.keys() for entity in entities]):
            filtered_dict[parent] = entities
    return filtered_dict


def get_distinct_colors(n: int) -> list:
    """

    :param n: number of distinct colors.
    :type n: ``int``
    :return: list with the hex code of the colors.
    :rtype: ``list``
    """
    colors_list = []
    l = 0.55
    s = 0.7
    for h in np.arange(0., 1., 1. / n):
        hex_code = colors.to_hex(hls_to_rgb(h, l, s))
        colors_list.append(hex_code)

    return colors_list


def create_dict_colors(parent_entity_dict: dict) -> dict:
    """
    Change the key value of the parent-entity dictionary to the color code.

    :param parent_entity_dict: dictionary with the relation parent - entities.
    :type parent_entity_dict: ``dict``
    :return: dictionary with the relation color-entities.
    :rtype: ``dict``
    """
    max_cluster = len(parent_entity_dict)
    color_dict = {}
    k = 0
    list_color = get_distinct_colors(max_cluster)
    for value in parent_entity_dict.values():
        code_hex = list_color[k]
        k += 1
        color_dict[code_hex] = value
    return color_dict


def save_word_cloud(full_path, word_cloud):
    word_cloud.to_file(full_path)

class SimpleGroupedColorFunc(object):
    def __init__(self, color_to_words):
        """

        :param color_to_words:
        """
        self.word_to_color = {word: color
                              for (color, words) in color_to_words.items()
                              for word in words}

    def __call__(self, word, **kwargs):
        """

        :param word:
        :param kwargs:
        :return:
        """
        return self.word_to_color[word]



import os
import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from TakeBlipInsightExtractor.outputs.eventhub_log_sender import \
    EventHubLogSender


class EmailSender:
    def __init__(self):
        self.sender = 'insightextractor.dataanalytics@gmail.com'
        self.username = self.sender
        self.__password = os.environ['EMAIL_PASSWORD']
        self.subject = 'Insight Extractor - Processamento Finalizado'
        self.host = 'smtp.gmail.com'
        self.port = 587

    @staticmethod
    def create_email_text(files_url_dict: dict, elapsed_minutes: float,
                          file_name: str) -> str:
        """
        Create the email corpus with the url of the files of the analysis.

        :param files_url_dict: Dictionary with the url of the files generated
        of the analysis
        :type files_url_dict: dict
        :param elapsed_minutes: Time elapsed to generate the analysis
        :type elapsed_minutes: float
        :param file_name: Name of the file send to analysis
        :type file_name: str
        :return: Email corpus to be send to the user
        :rtype: str
        """

        email_text = '''Para as frases enviadas no arquivo {}, foram 
        encontradas diferentes entidades, as quais foram agrupadas conforme 
        a imagem com a nuvem de entidades ({}). Nesse arquivo, as cores 
        indicam o grupo de entidades e o tamanho é proporcional a frequência 
        de vezes que essa entidade apareceu nas mensagens. O agrupamento das 
        entidades está no arquivo de hierarquia entidades ({}). 
        Similarmente, o arquivo json ({}) mostra tópicos relacionados às 
        entidades com hierarquia. Além disso é possível visualizar a 
        diferença de frequência dos 5 assuntos mais falados ({}). Mais 
        detalhes com as frases para cada entidade podem ser vistas no 
        arquivo csv que contém as informações para cada mensagem ({}). Tempo 
        de processamento: {:.2f} minutos '''.format(
            file_name,
            files_url_dict['word_cloud'],
            files_url_dict['entity_hierarchy'],
            files_url_dict['entity_hierarchy_dict'],
            files_url_dict['bar_chart'],
            files_url_dict['output_csv'],
            elapsed_minutes
        )
        return email_text

    def send_email(self, user_email: str, email_text: str,
                   logger: EventHubLogSender) -> None:
        """
        Send the e-mail to the user with the analysis link

        :param user_email: user e-mail to receive the analysis
        :type user_email: str
        :param email_text: e-mail corpus to be send to the user
        :type email_text: str
        :param logger: event hub logger
        :type logger: EventHubLogSender
        """
        self.__logger = logger
        self.__set_destination_emails(user_email, email_text)
        self.__connect_to_server()
        self.__dispatch_email(user_email)
        self.__disconnect_from_server()

    def __set_destination_emails(self, user_email: str,
                                 email_text: str) -> None:
        """
        Set the e-mail with the sender, the e-mail to receive, the subject and
        the corpus.

        :param user_email: user e-mail to receive the analysis
        :type user_email: str
        :param email_text: e-mail corpus to be send to the user
        :type email_text: str
        """

        try:
            self.__logger.log_message('DEBUG', 'Setting email addresses.')
            recipients = [user_email, 'analytics.dar@take.net']
            self.msg = MIMEMultipart()
            self.msg['From'] = self.sender
            self.msg['To'] = ', '.join(recipients)
            self.msg['Subject'] = self.subject
            self.msg.attach(MIMEText(email_text))
            self.__logger.log_message('DEBUG', 'Email addresses set.')
        except (Exception) as e:
            self.__logger.log_error_message('ERROR',
                                            'Error {} while setting email '
                                            'addresses!'.format(e))

    def __connect_to_server(self):
        """
        Connect to e-mail server.
        """
        try:
            self.__logger.log_message('DEBUG', 'Started server connection.')
            self.smtp = smtplib.SMTP(host=self.host, port=self.port)
            self.smtp.starttls()
            self.smtp.login(self.username, self.__password)
            self.__logger.log_message('DEBUG',
                                      'Email server connected successfully.')
        except (Exception) as e:
            self.__logger.log_error_message('ERROR',
                                            'Error {} connecting to email '
                                            'server!'.format(e))

    def __dispatch_email(self, user_email: str) -> None:
        """
        Send the e-mail with the analysis.

        :param user_email: user e-mail to receive the analysis
        :type user_email: str
        """
        try:
            self.__logger.log_message('DEBUG', 'Sending email.')
            self.smtp.sendmail(self.sender, user_email, self.msg.as_string())
            self.__logger.log_message('DEBUG', 'Email sent successfully.')
        except (Exception) as e:
            self.__logger.log_error_message('ERROR',
                                            'Error {} while sending '
                                            'email!'.format(e))

    def __disconnect_from_server(self):
        """
        Disconnect to e-mail server.
        """
        try:
            self.smtp.close()
            self.__logger.log_message('DEBUG',
                                      'Email server connection closed.')
        except (Exception) as e:
            self.__logger.log_error_message('ERROR',
                                            'Error {} while closing '
                                            'connection to email '
                                            'server!'.format(
                                                e))



import csv
from typing import Optional, List
from TakeBlipInsightExtractor.outputs.eventhub_log_sender import \
    EventHubLogSender
from TakeBlipMessageStructurer.predict import MessageStructurer
from take_text_preprocess.presentation import pre_process
from TakeBlipInsightExtractor.inputs.datalake_connector import DatabricksConnector

class DataGenerator:
    def __init__(self, data, separator):
        self.data = data
        self.separator = separator

    @classmethod
    def from_multipart_data_request(cls, part_data, encoding, separator):
        file_part_data = part_data.text.splitlines()
        return cls(file_part_data, separator)

    @classmethod
    def from_file_data(cls, path, encoding, separator):
        file_handler = open(path, encoding=encoding)
        return cls(file_handler, separator)

    def iterate_data(self):
        csv_reader = csv.reader(self.data, delimiter=self.separator)
        for row in csv_reader:
            yield row


class OutputFile:

    def __init__(self, input_data, encoding, separator, data_from='request',
                 logger=None, datalake_connector=None):
        """
        Parameters
        ----------
        input_data : str
            path to the csv file to be analyzed
        encoding : str
            file encoding
        separator : str
            csv file delimiter character
        data_from : str
            string to indicate if the file is from a request or local
        logger : EventHubLogSender
            event hub logger
        datalake_connector : DatabricksConnector
            datalake connector to receive messages from datalake
        """
        self.__messages_list = None
        self.input_data = input_data
        self.encoding = encoding
        self.separator = separator
        self.__logger = logger
        self.__datalake_connector = datalake_connector
        self.get_data = self.get_data_function(data_from)
        self.__set_header()

    def get_data_function(self, data_from):
        if data_from == 'request':
            return self.get_data_from_request
        elif data_from == 'file':
            return self.get_data_from_file
        elif data_from == 'datalake':
            return self.get_data_from_datalake
        else:
            raise Exception('Data origin not accepted')

    def get_data_from_file(self):
        return DataGenerator.from_file_data(
            path=self.input_data,
            encoding=self.encoding,
            separator=self.separator
        ).iterate_data()

    def get_data_from_request(self):
        return DataGenerator.from_multipart_data_request(
            part_data=self.input_data,
            encoding=self.encoding,
            separator=self.separator
        ).iterate_data()

    def get_data_from_datalake(self):
        messages = self.__datalake_connector.consume_data()
        return messages

    def __set_header(self):
        self.header = ['Mensagens', 'Tópicos', 'Assuntos']

    def parse_message_structurer_content(self, structured_messages_list: list) \
            -> (list, list):
        """Generate the column with the filtered messages and the entities

        Parameters
        ----------
        structured_messages_list : list
            List with structured message (with entities and filtered_message)
            of each row.

        Returns
        -------
        (list, list)
            List with the filtered message of each row, and list with entities
            of each row.
        """
        filtered_message_column = []
        entities_column = []
        for structured_message in structured_messages_list:
            filtered_message_column.append(
                structured_message['lowercase_filtered_message'])
            content = structured_message['content']
            if len(content) > 0:
                entities_list = [tag for tag in content
                                 if tag['type'] != 'postagging']
            else:
                entities_list = []
            entities_column.append(entities_list)
        return filtered_message_column, entities_column

    def __get_messages_list(self):
        return self.__messages_list

    def create_message_structurer_columns(self, batch_size: int,
                                          message_structurer: MessageStructurer,
                                          tags_to_remove: list,
                                          messages: Optional[List[str]]=None) -> (list, list):
        """
        Parameters
        ----------
        batch_size : int
            size of each batch to the prediction
        message_structurer : MessageStructurer
            message structurer model
        tags_to_remove : list
            list with tags to be removed in the message structurer

        Returns
        -------
        (list, list)
            List with the filtered message of each row, and list with entities
            of each row.

        Args:
            messages:
        """

        if messages:
            self.__messages_list = messages
            self.get_data = self.__get_messages_list

        ms_input_batch = [{'id': ind, 'sentence': pre_process(row[0],
                                                              ["URL", "EMAIL", "NUMBER", "CODE", "SYMBOL", "EMOJI"])}
                          for ind, row in enumerate(self.get_data())]

        self.log_message('DEBUG', 'Number of messages sent to be '
                                  'structured {}.'
                         .format(len(ms_input_batch)))
        self.log_message('DEBUG', 'Maximum message size {}'
                         .format(max(len(message['sentence'].split())
                                     for message in ms_input_batch)))

        structured_messages_list = sorted(
            message_structurer.structure_message_batch(
                batch_size=batch_size,
                sentences=ms_input_batch,
                tags_to_remove=tags_to_remove,
                use_pre_processing=False
            ),
            key=lambda content_dict: content_dict['id']
        )
        self.log_message('DEBUG', 'Messages structured with success.')
        filtered_message_column, entities_column = \
            self.parse_message_structurer_content(structured_messages_list)
        return filtered_message_column, entities_column

    def log_message(self, level, message):
        if self.__logger:
            self.__logger.log_message(level, message)

    @staticmethod
    def create_groups_column(entities_column: list,
                             parent_entity_dict: dict) -> list:
        """
        Parameters
        ----------
        entities_column : list
            list with entities of each row
        parent_entity_dict: dict
            dictionary with the relation parent - entities

        Returns
        -------
        list
            List with the name of the group of each entity of each row.
        """
        group_column = []
        entity_parent_dict = {child: parent for parent, children_lst in
                              parent_entity_dict.items() for child in
                              children_lst}

        for entities_list in entities_column:
            temp_list = []
            for entity_dict in entities_list:
                entity_name = entity_dict['lowercase_value']
                parent = entity_parent_dict[entity_name]
                temp_list.append(parent)
            group_column.append(temp_list)
        return group_column

    def create_messages_frequency_dict(self, entities_column: list) \
            -> dict:
        """
        Create a dictionary with the entities of a sentence as key and list
        with all sentences that has this entity, and their frequency as values.

        Parameters
        ----------
        entities_column : list
            list with entities of each row

        Returns
        -------
        dict
            Dictionary withe the entity as key and a list with a tuple of
            a sentence and the frequency as values.
        """
        messages_frequency_dict = {}
        entity_message_dict = {}
        for row in self.get_data():
            message = row[0]
            messages_frequency_dict[message] = messages_frequency_dict.get(
                message, 0) + 1

        for row, entities_list in zip(self.get_data(),
                                      entities_column):
            message = row[0]
            for entity_dict in entities_list:
                entity_name = entity_dict['lowercase_value']
                if entity_message_dict.get(entity_name) is not None:
                    entity_message_dict[entity_name].append((message,
                                                             messages_frequency_dict[
                                                                 message]))
                    entity_message_dict[entity_name] = list(
                        set(entity_message_dict[entity_name]))
                else:
                    entity_message_dict[entity_name] = [(
                        message, messages_frequency_dict[
                            message])]
        return entity_message_dict

    def __generate_output_file(self, entities_column: list, group_column: list):
        """
        Generate each output row

        Parameters
        ----------
        entities_column : list
            list with entities of each row
        group_column : list
            List with the name of the group of each entity of each row.
        """
        for message, entity_dict, group in zip(
                self.get_data(),
                entities_column,
                group_column):
            if len(entity_dict) > 0:
                entity_list = [entity['lowercase_value'] for entity in
                               entity_dict]
            else:
                entity_list = []
            yield message[0], entity_list, group

    def save_output_file(self, full_path: str,
                         entities_column: list,
                         group_column: list):
        """
        Save csv file with the original message, the filtered message, the
        entities found and the groups of the entities.

        Parameters
        ----------
        full_path : str
            String with the path to save the file
        entities_column : list
            list with entities of each row
        group_column : list
            List with the name of the group of each entity of each row.
        """
        with open(full_path, 'w', encoding='utf-8', newline='') as f:
            writer = csv.writer(f, delimiter=self.separator)
            if self.header:
                writer.writerow(self.header)
            for row in self.__generate_output_file(entities_column,
                                                   group_column):
                writer.writerow(row)



from TakeBlipInsightExtractor.clustering.verbs_list import LIST_VERBS

class Embedding:
    """Wraps methods for embedding creation and convertion.

    Attributes
    ----------
    embedding
        Word embedding.

    Methods
    -------
    create_entities_embeddings(entity_postagging_dict)
        Return a dictionary with the representation of the entities.
    """

    def __init__(self, embedding) -> None:
        """
        Parameters
        ----------
        embedding
            Word embedding.
        """
        self.embedding = embedding

    def create_entities_embeddings(self, entity_postagging_dict: dict) -> dict:
        """Return a dictionary with the representation of the entities.

        Parameters
        ----------
        entity_postagging_dict : dict
            Dictionary with the entity as key and POSTagging of the 
            entity as value.

        Returns
        -------
        dict
            Dictionary with the entity as key and the embedding 
            representation of the entity as value.
        """

        entities_embeddings_dict = {}
        for word in entity_postagging_dict.keys():
            if len(word.split()) > 1:
                tags = entity_postagging_dict[word]
                entities_embeddings_dict[word] = self.__convert_sent2vec(word,
                                                                         tags)
            else:
                entities_embeddings_dict[word] = self.__convert_word2vec(word)
        return entities_embeddings_dict

    def __convert_word2vec(self, word: str) -> list:
        """Get the embedding representation of a single word.

        Parameters
        ----------
        word : str
            Entity to get the embedding representation.

        Returns
        -------
        list
            Embedding representation.
        """
        return self.embedding[word].tolist()

    def __convert_sent2vec(self, words: str, postag: str) -> list:
        """Get the embedding representation of a compound word.

        Calculates a numerical representation of each word 
        inside words, which will be the average of them. 
        The method ignores prepositions or conjunctions.

        Parameters
        ----------
        words : str
            Entities to get the embedding representation.
        postag : str
            Description

        Returns
        -------
        list
            Embedding representation.
        """
        representation = []
        words_lst = words.split()
        postag_lst = postag.split()
        length = len(words_lst)
        for k in range(len(words_lst)):
            if postag_lst[k] in ['PREP', 'CONJ']:
                length = length - 1
            elif postag_lst[k] in ['VERB', 'PART']:
                if words_lst[k] in LIST_VERBS:
                    length = length - 1
                else:
                    representation.append(0.8 * self.embedding[words_lst[k]])
            else:
                representation.append(self.embedding[words_lst[k]])
        value = sum(representation)
        value = value / length
        return value.tolist()



from math import floor


def filter_entity(entities_column: list, entities_to_keep: set,
                  field_name: str) -> list:
    """Filter the entities a specific class

    Parameters
    ----------
    entities_column : list
        list with the entities of each row
    entities_to_keep : set
        set with the the match of a field to filter
    field_name : str
        string indicating which field should be used to filter the entities

    Returns
    -------
    list
        list with the entities filtered
    """

    filtered_entities_list = []
    for entities_list in entities_column:
        if len(entities_list) > 0:
            filtered_list = [entity_dict for entity_dict in entities_list
                             if entity_dict[field_name] in entities_to_keep]
        else:
            filtered_list = []
        filtered_entities_list.append(filtered_list)
    return filtered_entities_list


def filter_entity_type(entities_column: list) -> list:
    """Filter the entities to keep only the desired type
    Parameters
    ----------
    entities_column : list
        List with entities of each row

    Returns
    -------
    list
        List with the entities of each row, filtered to contained only the
        entities of specific class.
    """
    entity_list = filter_entity(
        entities_column=entities_column,
        entities_to_keep={'generic_entity', 'financial'},
        field_name='type')
    return entity_list


def filter_entity_frequency(entities_column: list,
                            entity_frequency_dict: dict) -> list:
    """Filter the entities based on the frequency

    Parameters
    ----------
    entities_column : list
        List with entities of each row
    entity_frequency_dict : dict
         Dictionary with entities as keys and frequency as values,
         filtered by the frequency

    Returns
    -------
    list
        List with the entities of each row, filtered to the most frequently
    """
    entity_list = filter_entity(
        entities_column=entities_column,
        entities_to_keep=set(entity_frequency_dict.keys()),
        field_name='lowercase_value')
    return entity_list


def create_postagging_dict(entities_column: list) -> dict:
    """Create dictionary with the postag of each entity

    Parameters
    ---------
    entities_column : list
        list with entities of each row

    Returns
    -------
    dict
        Dictionary with key as the entities and values with the postag of the
        entities.
    """
    entity_postagging_dict = {
        entity_dict['lowercase_value']: entity_dict['postags']
        for entities_list in entities_column for entity_dict in entities_list}
    return entity_postagging_dict


def create_entity_frequency_dict(entities_column: list,
                                 percentage_threshold: float) -> dict:
    """
    Create the dictionary with the frequency of each entities and filtered
    using a certain frequency percentile

    Parameters
    ----------
    entities_column : list
        list with the entities of each row
    percentage_threshold : float
        percentile of frequency to filter

    Returns
    -------
    dict
        Dictionary with entities as keys and frequency as values
    """

    frequency = dict()
    for entities_list in entities_column:
        entities = map(dict, set(tuple(entity_dict.items()) for
                                 entity_dict in entities_list))
        for entity_dict in entities:
            entity_name = entity_dict['lowercase_value']
            frequency[entity_name] = frequency.get(entity_name, 0) + 1
    sorted_frequencies_lst = sorted(frequency.values(), reverse=True)
    number_items = floor(percentage_threshold * len(frequency))
    threshold_frequency = sorted_frequencies_lst[:number_items][-1]
    delete_low_frequency_entities(threshold_frequency, frequency)
    return frequency


def delete_low_frequency_entities(threshold_frequency, frequency_dict):
    frequency_dict_copy = frequency_dict.copy()

    if threshold_frequency == 1:
        threshold_frequency = 2

    for entity, frequency in frequency_dict_copy.items():
        if frequency < threshold_frequency:
            del frequency_dict[entity]



import numpy as np

from sklearn.cluster import AgglomerativeClustering
from sklearn.metrics.pairwise import cosine_distances


class EntityCluster:
    """Wraps methods for entities clustering.

    Attributes
    ----------
    sim_threshold : float
        Clustering similarity threshold.

    Methods
    -------
    cluster_entities(entities_embedding)
        Clusters a group of entities.

    # TODO: add public methods
    """

    def __init__(self, sim_threshold: float = 0.65):
        """
        Parameters
        ----------
        sim_threshold : float, optional
            Similarity threshold;
        """
        self.sim_threshold = sim_threshold

    def cluster_entities(self, entities_embedding: dict) -> dict:
        """Clusters a group of entities.

        Parameters
        ----------
        entities_embedding : dict
            Entities and their embedding representation.

        Returns
        -------
        dict
            Entity (key) to cluster number (value) mapping.
        """
        entity_cluster_dict = {}
        entities_array = np.array(list(entities_embedding.values()))
        thr = self.__calculate_threshold(entities_array)
        agglomerative = AgglomerativeClustering(n_clusters=None,
                                                linkage='complete',
                                                affinity='cosine',
                                                distance_threshold=thr
                                                ).fit(entities_array)

        keys = list(entities_embedding.keys())
        for k in range(len(keys)):
            key = keys[k]
            entity_cluster_dict[key] = agglomerative.labels_[k]
        return entity_cluster_dict

    def __calculate_threshold(self, entities_embeddig: np.array) -> float:
        """Calculates distance threshold.

        Calculates the distance threshold over the similarity matrix of a
        dataset, using a minimum value of similarity between the clusters.

        Parameters
        ----------
        entities_embeddig : np.array
            Entities and their embedding representation.

        Returns
        -------
        float
            Threshold of distance for a level of similarity.
        """
        cos = cosine_distances(entities_embeddig)
        return (1 - self.sim_threshold) * cos.max()



import json
import uuid
import socket
import logging
from pytz import timezone
from datetime import datetime
from azure.eventhub import EventData
from azure.eventhub import EventHubProducerClient


class EventHubLogSender():
    """Sends logs to EventHub.

     Methods:
      * log_message - Sends log to EventHub.
      * log_error_message - Logs error messages and sends it to EventHub.
      * generate_json_log - Generates log in JSON format.
      * generate_json_log_to_print - Generates log in printable format.
    """

    def __init__(
            self,
            application_name: str,
            user_email: str,
            bot_name: str,
            file_name: str,
            correlation_id: str,
            connection_string: str = None,
            eventhub_name: str = None) -> None:
        """
        :param application_name: Running application name.
        :type application_name: `str`
        :param user_email: Email from the user sending the request.
        :type user_email: `str`
        :param bot_name: Name of the bot being analyzed. 
        :type bot_name: `str`
        :param file_name: Name of the file being analyzed.
        :type file_name: `str`
        :param correlation_id: Execution unique id.
        :type correlation_id: `str` of `uuid`
        :param connection_string: Eventhub's connection string.
        :type connection_string: `str`
        :param eventhub_name: Name of the log event hub.
        :type eventhub_name: `str`
        """
        self.__connection_string = connection_string
        self.__eventhub_name = eventhub_name
        self.__application_name = application_name
        self.__user_email = user_email
        self.__bot_name = bot_name
        self.__file_name = file_name
        self.__correlation_id = correlation_id
        self.__timezone = timezone('Brazil/East')
        self.__connect_to_eventhub()
        self.__generate_level_dict()
        logging.getLogger("uamqp").setLevel(logging.CRITICAL)
        logging.getLogger("azure.eventhub").setLevel(logging.CRITICAL)

    def __connect_to_eventhub(self) -> None:
        if self.__eventhub_is_initialized():
            self.__sender = EventHubProducerClient.from_connection_string(
                conn_str=self.__connection_string,
                eventhub_name=self.__eventhub_name)
            self.__event_batch = self.__sender.create_batch()

    def __eventhub_is_initialized(self) -> bool:
        return self.__eventhub_name and self.__connection_string

    def __generate_level_dict(self) -> None:
        self.__level_dict = {
            'DEBUG': 10,
            'INFO': 20,
            'WARNING': 30,
            'ERROR': 40,
            'CRITICAL': 50
        }

    def log_message(self, level: str, message: str) -> None:
        """Sends log to EventHub.

        :param level: Criticity level. 
            Can either be of "DEBUG", "INFO", "WARNING", 
            "ERROR" or "CRITICAL". 
        :type level: `str`
        :param message: Message informing the event.
        :type message: `str`
        """
        self.__connect_to_eventhub()
        self.__send_message_to_eventhub(level, message)

    def log_error_message(self, level: str, message: str) -> None:
        """Logs error messages and sends it to EventHub.

        :param level: Criticity level. 
            Can either be of "DEBUG", "INFO", "WARNING", 
            "ERROR" or "CRITICAL". 
        :type level: `str`
        :param message: Message informing the event.
        :type message: `str`
        """
        logging.error(message)
        self.__connect_to_eventhub()
        self.__send_message_to_eventhub(level, message)

    def __send_message_to_eventhub(self, level: str, message: str) -> None:
        timestamp = str(datetime.now(self.__timezone))

        self.generate_json_log(timestamp, level, message)
        print(self.generate_json_log_to_print(timestamp, level, message))

        if self.__eventhub_is_initialized():
            with self.__sender:
                self.__event_batch.add(EventData(self.json_log))
                self.__sender.send_batch(self.__event_batch)

    def generate_json_log(
            self,
            timestamp: str,
            level: str,
            message: str) -> None:
        """Generates log in JSON format.

        :param timestamp: When the log happened.
        :type timestamp: `str`
        :param level: Criticity level. 
            Can either be of "DEBUG", "INFO", "WARNING", 
            "ERROR" or "CRITICAL". 
        :type level: `str`
        :param message: Message informing the event.
        :type message: `str`
        """
        log_id = str(
            uuid.uuid3(
                uuid.NAMESPACE_DNS,
                self.__user_email + self.__bot_name + timestamp))

        extras = {
            'UserEmail': self.__user_email,
            'BotName': self.__bot_name,
            'FileName': self.__file_name,
            'HostName': socket.gethostname(),
        }

        self.json_log = json.dumps({
            'StorageDateBR': timestamp,
            'CorrelationId': self.__correlation_id,
            'Id': log_id,
            'Application': self.__application_name,
            'Level': level,
            'Message': message,
            'Extras': extras
        })

    def generate_json_log_to_print(
            self,
            timestamp: str,
            level: str,
            message: str) -> str:
        """Generates log in printable format.

        :param timestamp: When the log happened.
        :type timestamp: `str`
        :param level: Criticity level. 
            Can either be of "DEBUG", "INFO", "WARNING", 
            "ERROR" or "CRITICAL". 
        :type level: `str`
        :param message: Message informing the event.
        :type message: `str`
        :return: Log in printable format
        :rtype: `str`
        """

        return '{} : {} : {} : {} : {} : {} : {} : {}'.format(
            level,
            timestamp,
            self.__correlation_id,
            self.__application_name,
            socket.gethostname(),
            self.__user_email,
            self.__bot_name,
            message)



import datetime
import base64
from typing import Optional, List

from TakeBlipInsightExtractor.inputs import models_loader
from TakeBlipInsightExtractor.processing import entity_filter
from TakeBlipInsightExtractor.clustering import embedding, entity_cluster
from TakeBlipInsightExtractor.visualization import entity_hierarchy, \
    word_cloud, csv_file
from TakeBlipInsightExtractor.outputs.eventhub_log_sender import \
    EventHubLogSender
from TakeBlipMessageStructurer.predict import MessageStructurer
from TakeBlipInsightExtractor.inputs.datalake_connector import DatabricksConnector

class InsightExtractor:

    def __init__(self,
                 input_data: str,
                 separator: str,
                 similarity_threshold: float,
                 embedding_path: str,
                 postagging_model_path: str,
                 postagging_label_path: str,
                 ner_model_path: str,
                 ner_label_path: str,
                 user_email: str,
                 bot_name: str,
                 use_neighborhood: bool,
                 logger: EventHubLogSender,
                 datalake_connector: DatabricksConnector=None):
        """
        Parameters
        ----------
        input_data : str
            path to the csv file to be analyzed
        separator : str
            csv file delimiter character
        similarity_threshold : float
            threshold of similarity between clusters
        embedding_path : str
            path to embedding model
        postagging_model_path : str
            path to the postagging model
        postagging_label_path : str
            path to the labels dictionary of postag model
        ner_model_path : str
            path to the ner model
        ner_label_path : str
            path to the labels dictionary of ner model
        user_email : str
            users e-mail where they want to receive the analyzes
        bot_name: str
            bot ID
        use_neighborhood : bool
            use or not the neighborhood analysis.
        logger : EventHubLogSender
            event hub logger
        datalake_connector : DatabricksConnector
            datalake connector to receive messages from datalake
        """
        self.__messages = None
        self.__logger = logger
        self.__datalake_connector = datalake_connector
        self.__set_entity_cluster(similarity_threshold)
        self.__set_message_structurer(
            postagging_model_path=postagging_model_path,
            postagging_label_path=postagging_label_path,
            embedding_path=embedding_path,
            ner_model_path=ner_model_path,
            ner_label_path=ner_label_path,
            use_neighborhood=use_neighborhood
        )
        self.__set_output_csv_file(input_data, separator, self.__logger)
        self.__set_output_files_path_dict(user_email, bot_name)
        self.__set_embedding()

    def __set_entity_cluster(self, similarity_threshold: float) -> None:
        """Start clustering method

        Parameters
        ----------
        similarity_threshold : float
            threshold of similarity between clusters
        """
        self.__logger.log_message('DEBUG', 'Creating cluster layer.')
        self.entity_cluster = entity_cluster.EntityCluster(
            similarity_threshold)
        self.__logger.log_message('DEBUG', 'Cluster layer created.')

    def __set_message_structurer(self, postagging_model_path: str,
                                 postagging_label_path: str,
                                 embedding_path: str,
                                 ner_model_path: str,
                                 ner_label_path: str,
                                 use_neighborhood: bool) -> None:
        """ Load models and start message structurer

        Parameters
        ----------
        postagging_model_path : str
            path to the postagging model
        postagging_label_path : str
            path to the labels dictionary of postag model
        embedding_path : str
            path to embedding model
        ner_model_path : str
            path to the ner model
        ner_label_path : str
            path to the labels dictionary of ner model
        use_neighborhood : bool
            use or not the neighborhood analysis.
        """
        self.__logger.log_message('DEBUG', 'Setting Message Structurer.')
        self.__logger.log_message('DEBUG', 'Started reading the PosTagging '
                                           'model.')
        postag_predictor = models_loader.load_postagging_predictor(
            postagging_model_path=postagging_model_path,
            postagging_label_path=postagging_label_path,
            embedding_path=embedding_path
        )
        self.__logger.log_message('DEBUG', 'Finished reading the PosTagging '
                                           'model.')

        self.__logger.log_message('DEBUG', 'Started reading the NER model.')
        ner_predictor = models_loader.load_ner_predictor(
            ner_model_path=ner_model_path,
            ner_label_path=ner_label_path,
            postag_predictor=postag_predictor
        )
        self.__logger.log_message('DEBUG', 'Finished reading the NER model.')

        self.__logger.log_message('DEBUG', 'Started Message Structurer '
                                           'creation.')
        self.tags_to_remove = ['INT', 'ART', 'PRON', 'SIMB', 'PON', 'CONJ']
        self.message_structurer = MessageStructurer(
            ner_model=ner_predictor,
            use_neighborhood=use_neighborhood
        )
        self.__logger.log_message('DEBUG', 'Finished Message Structurer '
                                           'creation.')
        self.__logger.log_message('DEBUG', 'Message Structurer set.')

    def __set_output_csv_file(self, input_data: str, separator: str,
                              logger: EventHubLogSender = None) -> None:
        """Set the csv output file and start reading the input file
        Parameters
        ----------
        input_data : str
            path to the csv file to be analyzed
        separator : str
            csv file delimiter character
        logger : EventHubLogSender
            event hub logger
        """
        self.__logger.log_message('DEBUG', 'Setting output csv file.')
        if self.__datalake_connector:
            data_from = 'datalake'
        else:
            data_from = 'request' if type(input_data) != str else 'file'
        self.output_file = csv_file.OutputFile(
            input_data=input_data,
            encoding='utf-8',
            separator=separator,
            data_from=data_from,
            logger=logger,
            datalake_connector=self.__datalake_connector
        )
        self.__logger.log_message('DEBUG', 'Output csv file set.')

    def __set_output_files_path_dict(self, user_email: str,
                                     bot_name: str) -> None:
        """Set the dictionary with the path of all outputs
        Parameters
        ----------
        user_email : str
            users e-mail where they want to receive the analyzes
        bot_name: str
            bot ID
        """
        self.__logger.log_message('DEBUG', 'Setting Insight Extractor output '
                                           'files names.')
        date = datetime.datetime.now()

        code = '{}_{}_{}{}{}_{}{}{}'.format(
            user_email,
            bot_name,
            date.year,
            date.month,
            date.day,
            date.hour,
            date.minute,
            date.second
        )

        code = base64.b64encode(code.encode('ascii')).decode('ascii')

        self.output_files_dict = {
            'word_cloud': 'word_cloud_{}.png'.format(code),
            'entity_hierarchy': 'entity_hierarchy_{}.html'.format(code),
            'output_csv': 'output_{}.csv'.format(code),
            'entity_hierarchy_dict': 'entity_hierarchy_dict_{}.json'
                .format(code),
            'bar_chart': 'bar_chart_{}.png'.format(code)
        }
        self.__logger.log_message('DEBUG', 'Insight Extractor output files '
                                           'names set.')

    def __set_embedding(self):
        """Set embedding to be used on the clustering algorithm
        """
        self.__logger.log_message('DEBUG', 'Setting embedding.')
        self.embedding = embedding.Embedding(
            self.message_structurer.ner_model.fasttext)
        self.__logger.log_message('DEBUG', 'Embedding set.')

    def predict(self,
                percentage_threshold: float,
                node_messages_examples: int,
                batch_size: int,
                messages: Optional[List[str]] = None) -> None:
        """Generate clusters with the topics
        Parameters
        ---------
        percentage_threshold : float
            percentile of frequency to filter.
        node_messages_examples : int
            number of examples for each entity
        batch_size : int
            size of each batch to the prediction
        messges : tp.List[str]
            list of messages to analyze
        """
        self.__messages = messages
        self.__logger.log_message('DEBUG', 'Predict started.')
        self.__structure_messages(batch_size, self.__messages)
        self.__filter_messages(percentage_threshold)
        self.__generate_entities_embedding()
        self.__cluster_messages()
        self.__filter_messages_by_frequency()
        self.__generate_word_cloud()
        self.__generate_entity_hierarchy(node_messages_examples)
        self.__save_files_locally()
        self.__logger.log_message('DEBUG', 'Insight Extractor predict '
                                           'finished.')

    def __structure_messages(self, batch_size: int, messages:Optional[List[str]]) -> None:
        """
        Parameters
        ----------
        batch_size : int
            size of each batch to the prediction
        """
        self.__logger.log_message('DEBUG', 'Creating Message Structurer '
                                           'column.')
        self.__filtered_message_column, self.__entities_column = self.output_file. \
            create_message_structurer_columns(batch_size,
                                              self.message_structurer,
                                              self.tags_to_remove,
                                              messages)
        self.__logger.log_message('DEBUG', 'Message Structurer column '
                                           'created.')

    def __cluster_messages(self) -> None:
        """
        Clusterize the entities, and create a dictionary with the entities and
        the cluster label; one dictionary with the cluster name and the entities
        in; and the column with the cluster of each sentence.
        """
        self.__logger.log_message('DEBUG', 'Clustering started.')
        self.__create_entity_cluster_dict()
        self.__create_parent_entity_dict()
        self.__create_group_column()
        self.__logger.log_message('DEBUG', 'Clustering finished.')

    def __create_entity_cluster_dict(self) -> None:
        """
        Clusterize the entities and generate the dictionary with the entities
        as keys and the cluster label as values.
        """
        self.__logger.log_message('DEBUG', 'Creating entity cluster dict.')
        self.__entity_cluster_dict = self.entity_cluster.cluster_entities(
            self.__entities_embedding)
        self.__logger.log_message('DEBUG', 'Entity cluster dict created.')

    def __create_parent_entity_dict(self) -> None:
        """
        Create a dictionary with the cluster name (the most frequenty entity)
        as key and all entities of the cluster as values.
        """
        self.__logger.log_message('DEBUG', 'Creating parent entity dict.')
        self.__parent_entity_dict = entity_hierarchy.create_parent_dict(
            self.__entity_frequency_dict, self.__entity_cluster_dict)
        self.__logger.log_message('DEBUG', 'Parent entity dict created.')

    def __create_group_column(self) -> None:
        """
        Create a list with the groups of each row.
        """
        self.__logger.log_message('DEBUG', 'Creating group column.')
        self.__group_column = self.output_file.create_groups_column(
            self.__entities_column, self.__parent_entity_dict)
        self.__logger.log_message('DEBUG', 'Group column created.')

    def __filter_messages_by_frequency(self) -> None:
        """
        Create a dictionary with the entities of a sentence as key and list
        with all sentences that has this entity, and their frequency as values.
        """
        self.__logger.log_message('DEBUG', 'Creating filtered messages '
                                           'frequency dict.')
        self.__filtered_messages_freq_dict = \
            self.output_file.create_messages_frequency_dict(
                self.__entities_column)
        self.__logger.log_message('DEBUG', 'Filtered messages frequency '
                                           'dict created.')

    def __generate_entities_embedding(self):
        """
        Generate the dictionary with the entity as key and the embedding
        representation of the entity as value.
        """
        self.__logger.log_message('DEBUG', 'Creating entities embedding.')
        self.__entities_embedding = self.embedding.create_entities_embeddings(
            self.__entity_postagging_dict)
        self.__logger.log_message('DEBUG', 'Entities embedding created.')

    def __filter_messages(self, percentage_threshold: float):
        """
        Parameters
        ---------
        percentage_threshold : float
            percentile of frequency to filter
        """
        self.__logger.log_message('DEBUG', 'Creating entities column.')
        self.__entities_column = entity_filter.filter_entity_type(
            self.__entities_column)
        self.__logger.log_message('DEBUG', 'Entities column created.')

        self.__logger.log_message('DEBUG', 'Creating entity frequency dict.')
        self.__entity_frequency_dict = entity_filter.create_entity_frequency_dict(
            self.__entities_column, percentage_threshold)
        self.__logger.log_message('DEBUG', 'Entity frequency dict created.')
        self.__logger.log_message('DEBUG', '{} entities were found.'
                                  .format(len(self.__entity_frequency_dict)))

        self.__logger.log_message('DEBUG', 'Filtering messages by entities.')
        self.__entities_column = entity_filter.filter_entity_frequency(
            self.__entities_column, self.__entity_frequency_dict)
        self.__logger.log_message('DEBUG', 'Messages filtered.')

        self.__logger.log_message('DEBUG', 'Creating entity postagging dict.')
        self.__entity_postagging_dict = entity_filter.create_postagging_dict(
            self.__entities_column)
        self.__logger.log_message('DEBUG', 'Entity postagging dict created.')

    def __generate_word_cloud(self):
        """
        Generate word cloud visualization
        """
        self.__logger.log_message('DEBUG', 'Creating wordcloud image.')
        self.__word_cloud_image = \
            word_cloud.create_word_cloud(self.__entity_frequency_dict,
                                         self.__parent_entity_dict)
        self.__logger.log_message('DEBUG', 'Wordcloud image created.')

    def __generate_entity_hierarchy(self, node_messages_examples: int) -> None:
        """
        Generate entity hierarchy visualization
        Parameters
        ----------
        node_messages_examples : int
            number of examples for each entity
        """
        self.__logger.log_message('DEBUG', 'Creating entity hierarchy html '
                                           'file.')
        self.__entity_hierarchy_html = entity_hierarchy.create_entity_hierarchy(
            self.__filtered_messages_freq_dict, self.__parent_entity_dict,
            node_messages_examples)
        self.__logger.log_message('DEBUG', 'Entity hierarchy html file '
                                           'created.')

        self.__logger.log_message('DEBUG', 'Creating entity hierarchy dict '
                                           'file.')
        self.__hierarchy_freq_dict, self.__parent_freq_dict = entity_hierarchy.create_hierarchy_frequency_dict(
            self.__entity_frequency_dict, self.__parent_entity_dict)
        self.__logger.log_message('DEBUG', 'Entity hierarchy dict file '
                                           'created.')

    def __save_files_locally(self):
        self.__logger.log_message('DEBUG', 'Saving output file.')
        self.output_file.save_output_file(self.output_files_dict['output_csv'],
                                          self.__entities_column,
                                          self.__group_column)
        self.__logger.log_message('DEBUG', 'Output file saved.')

        self.__logger.log_message('DEBUG', 'Saving wordcloud image.')
        word_cloud.save_word_cloud(self.output_files_dict['word_cloud'],
                                   self.__word_cloud_image)
        self.__logger.log_message('DEBUG', 'Wordcloud image saved.')

        self.__logger.log_message('DEBUG', 'Saving entity hierarchy html '
                                           'file.')
        entity_hierarchy.save_entity_hierarchy(
            self.output_files_dict['entity_hierarchy'],
            self.__entity_hierarchy_html)
        self.__logger.log_message('DEBUG', 'Entity hierarchy html file saved.')

        self.__logger.log_message('DEBUG', 'Saving entity hierarchy dict '
                                           'file.')
        entity_hierarchy.save_entity_hierarchy_dict(
            self.output_files_dict['entity_hierarchy_dict'],
            self.__hierarchy_freq_dict)
        self.__logger.log_message('DEBUG', 'Entity hierarchy dict file saved.')

        self.__logger.log_message('DEBUG', 'Saving top 5 figure.')
        entity_hierarchy.save_top_plot(self.output_files_dict['bar_chart'],
                                       self.__parent_freq_dict)
        self.__logger.log_message('DEBUG', 'Top 5 figure saved.')



import os
import re
import ast


def get_ie_static_parameters():
    return {
        'node_messages_examples': get_node_messages_examples(),
        'similarity_threshold': get_similarity_threshold(),
        'percentage_threshold': get_percentage_threshold(),
        'batch_size': get_batch_size(),
        'chunk_size': get_chunk_size(),
        'embedding_path': get_embedding_path(),
        'postagging_model_path': get_postagging_model_path(),
        'postagging_label_path': get_postagging_label_path(),
        'ner_model_path': get_ner_model_path(),
        'ner_label_path': get_ner_label_path(),
        'eventhub_name': get_eventhub_name(),
        'application_name': get_application_name(),
        'server_host_name': get_server_host_name(),
        'http_path': get_http_path(),
        'access_token': get_access_token(),
        'message_column': get_message_column(),
        'storage_date_column': get_storage_date_column(),
        'identity_column': get_identity_column(),
        'data_table': get_data_table(),
        'message_id_column': get_message_id_column(),
        'subject_data_table': get_subject_data_table(),
        'subject_bot_name_column': get_subject_bot_name_column(),
        'subject_message_id_column': get_subject_message_id_column(),
        'use_neighborhood': get_use_neighborhood()
    }


def get_embedding_path():
    return get_azureml_file_path(
        env_model_name='EMBEDDING_VERSION',
        env_full_name='EMBEDDING_FULL_NAME',
        env_registry_name='EMBEDDING_REGISTRY_NAME'
    )


def get_postagging_model_path():
    return get_azureml_file_path(
        env_model_name='POSTAGGING_VERSION',
        env_full_name='POSTAGGING_FULL_NAME',
        env_registry_name='POSTAGGING_REGISTRY_NAME'
    )


def get_postagging_label_path():
    return get_azureml_file_path(
        env_model_name='POSTAGGING_LABEL_VERSION',
        env_full_name='POSTAGGING_LABEL_FULL_NAME',
        env_registry_name='POSTAGGING_LABEL_REGISTRY_NAME'
    )


def get_ner_model_path():
    return get_azureml_file_path(
        env_model_name='NER_VERSION',
        env_full_name='NER_FULL_NAME',
        env_registry_name='NER_REGISTRY_NAME'
    )


def get_ner_label_path():
    return get_azureml_file_path(
        env_model_name='NER_LABEL_VERSION',
        env_full_name='NER_LABEL_FULL_NAME',
        env_registry_name='NER_LABEL_REGISTRY_NAME'
    )


def get_azureml_file_path(env_model_name, env_full_name, env_registry_name):
    azureml_dir = os.environ.get('AZUREML_MODEL_DIR')
    file_version = os.environ.get(env_model_name)
    file_full_name = os.environ.get(env_full_name)
    file_registry_name = os.environ.get(env_registry_name)
    return os.path.join(
        azureml_dir,
        file_registry_name,
        file_version,
        file_full_name
    )


def get_node_messages_examples():
    return int(os.environ.get('NUMBER_NODE_MESSAGES_EXAMPLES', 6))


def get_similarity_threshold():
    return float(os.environ.get('SIMILARITY_THRESHOLD', 0.65))


def get_percentage_threshold():
    return float(os.environ.get('PERCENTAGE_THRESHOLD', 0.7))


def get_batch_size():
    return int(os.environ.get('BATCH_SIZE', 50))


def get_chunk_size():
    return int(os.environ.get('CHUNK_SIZE', 1024))


def get_environment():
    return os.environ.get('ENVIRONMENT', 'prod')


def get_eventhub_name():
    return os.environ.get('EVENTHUB_NAME', None)


def get_application_name():
    return os.environ.get('APPLICATION_NAME', None)


def get_server_host_name():
    return os.environ.get('SERVER_HOSTNAME', None)


def get_http_path():
    return os.environ.get('HTTP_PATH', None)


def get_access_token():
    return os.environ.get('ACCESS_TOKEN', None)


def get_message_column():
    return os.environ.get('MESSAGE_COLUMN', None)


def get_storage_date_column():
    return os.environ.get('STORAGE_DATE_COLUMN', None)


def get_identity_column():
    return os.environ.get('IDENTITY_COLUMN', None)


def get_data_table():
    return os.environ.get('DATA_TABLE', None)


def get_message_id_column():
    return os.environ.get('MESSAGE_ID_COLUMN', None)


def get_subject_data_table():
    return os.environ.get('SUBJECT_DATA_TABLE', None)


def get_subject_bot_name_column():
    return os.environ.get('SUBJECT_BOT_NAME_COLUMN', None)


def get_subject_message_id_column():
    return os.environ.get('SUBJECT_MESSAGE_ID_COLUMN', None)

def get_use_neighborhood():
    return os.environ.get('USE_NEIGHBORHOOD', True)

def get_parameters_for_multipart_request(multipart_data):
    parameters = {
        'separator': '|',
        'file_name': 'file.csv',
        'bot_name': None,
        'user_email': None,
        'file': None,
    }
    for part in multipart_data.parts:
        disposition = part.headers[b'Content-Disposition'].decode('utf-8')
        parameter_name = re.search('name="(.*?)"', disposition.split(';')[1]).group(1)
        if parameter_name == 'bot_name':
            parameters['bot_name'] = part.text
        elif parameter_name == 'user_email':
            parameters['user_email'] = part.text
        elif parameter_name == 'separator':
            parameters['separator'] = part.text
        elif parameter_name == 'file':
            parameters['file'] = part
            parameters['file_name'] = re.search('filename="(.*?)"', disposition.split(';')[2]).group(1)
    return parameters


def check_parameters(parameters):
    for parameter_name in parameters:
        if not parameter_name:
            raise ValueError('Missing {} parameter in form data request'.format(parameter_name))


def get_parameters_for_json_request(request):
    parameters = {
        'bot_name': None,
        'user_email': None,
        'start_date': None,
        'end_date': None,
        'bot_identity': None,
        'separator': '|',
        'file_name': 'file.csv',
        'file': None,
    }

    dict_str = request.decode("UTF-8")
    request_data = ast.literal_eval(dict_str)

    for parameter_name in request_data.keys():
        if parameter_name == 'bot_name':
            parameters[parameter_name] = request_data[parameter_name]
        elif parameter_name == 'user_email':
            parameters[parameter_name] = request_data[parameter_name]
        elif parameter_name == 'start_date':
            parameters[parameter_name] = request_data[parameter_name]
        elif parameter_name == 'end_date':
            parameters[parameter_name] = request_data[parameter_name]
        elif parameter_name == 'bot_identity':
            parameters[parameter_name] = request_data[parameter_name]

    return parameters



import os
import re
import uuid
import logging
import datetime
from azure.core.exceptions import ResourceExistsError
from azure.storage.blob import BlobServiceClient, BlobClient, ContainerClient, BlobBlock

class Storage:
    def __init__(self, environment):
        self.environment = environment
        self.__set_blob_service_client(environment)

    def __get_connection_string(self):
        env_variable_name = 'HMG_STORAGE_CONNECTION_STRING' if self.environment == 'hmg' else 'STORAGE_CONNECTION_STRING'
        connection_string = os.environ[env_variable_name]
        return connection_string
        
    def __set_blob_service_client(self, environment):
        self.blob_service_client = BlobServiceClient.from_connection_string(
            conn_str=self.__get_connection_string()
        )
        
    def __get_blob_client(self, user_container_name, directory_name, file_path):
        blob_client = BlobClient.from_connection_string(
            conn_str=self.__get_connection_string(),
            container_name=user_container_name,
            blob_name=os.path.join(directory_name, file_path)
        )
        return blob_client

    def __standard_characters(self, value):
        return value if value >= 10 else '0' + str(value)
    def __get_directory_name(self, bot_name):
        date = datetime.datetime.now() 
        directory_name = '{}_{}{}{}_{}{}{}'.format(
            bot_name,
            date.year,
            self.__standard_characters(date.month),
            self.__standard_characters(date.day),
            date.hour,
            date.minute,
            date.second
        )
        return directory_name 
    
    def __upload_file(self, file_path, blob_client, chunk_size, logger):
        logging.disable(logging.INFO)
        block_list = []
        try:
            logger.log_message('DEBUG','Uploading {}'.format(file_path))
            with open(file_path, "rb") as f:
                while True:
                    read_data = f.read(chunk_size)
                    if not read_data:
                        break
                    blk_id = str(uuid.uuid4())
                    blob_client.stage_block(block_id=blk_id,data=read_data) 
                    block_list.append(BlobBlock(block_id=blk_id))
            blob_client.commit_block_list(block_list)
            logger.log_message('DEBUG','{} uploaded'.format(file_path))
        except Exception as e:
            logger.log_error_message('ERROR', 'Error {} while uploading file {}'.format(e, file_path))
            
    def create_user_container(self, user_email, logger):
        user_container_name = re.search('(.*?)@take.net', user_email).group(1).replace('.','')
        try:
            self.blob_service_client.create_container(
                name=user_container_name,
                public_access='Blob'
            )
        except ResourceExistsError:
            logger.log_message('DEBUG','Container for user {} already exists'.format(user_container_name))
        except Exception as e:
            logger.log_error_message('ERROR', 'Error {} while creating user container.'.format(e))
        else:
            logger.log_message('DEBUG','Created container for user {}'.format(user_container_name))
        finally:
            return user_container_name

    def save_files_to_storage(self, save_files_dict, user_container_name, bot_name, chunk_size, logger):
        files_url_dict = {}
        directory_name = self.__get_directory_name(bot_name)
        for file_name, file_path in save_files_dict.items():
            blob_client = self.__get_blob_client(user_container_name, directory_name, file_path)
            files_url_dict[file_name] = blob_client.url
            self.__upload_file(file_path, blob_client, chunk_size, logger)
        return files_url_dict


import setuptools

with open("README_PACKAGE.md", "r") as fh:
    long_description = fh.read()

with open("requirements.txt") as f:
    requirements = f.read().splitlines()

setuptools.setup(
    name="insight-extractor-package",
    version="0.0.5",
    author="Research and Innovation",
    author_email="insightextractor.dataanalytics@gmail.com",
    credits=['Renan Santos'],
    keywords='insight extractor',
    description="Insight Extractor Package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    install_requires=requirements,
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import operator
import json
import typing as tp
import plotly.express as px


def create_parent_dict(entity_frequency_dict: dict,
                       entity_cluster_dict: dict) -> dict:
    """Create the dictionaries for entity - cluster name

    Receive one dictionary with the frequency of each entity (entity as key
    and frequency as value) and one dictionary with the cluster id (entity
    as key and cluster id as values). And create two dictionaries:

    * Entities_parent_dictionary: dictionary with the relation of entity
    and cluster name (parent). Entity as key and cluster name as value.

    * Parent_entities_dictionary: dictionary with the relation of cluster
    name(parent) and all entities in the cluster. Cluster name as key and
    list with all entities as value.

    Parameters
    ----------
    entity_frequency_dict: dict
       dictionary entity as key and frequency as value.
    entity_cluster_dict: dict
        dictionary entity as key and cluster id as value.

    Returns
    -------
    dict
        Dictionary with the relation parent - entities
    """
    parent_entity_dict = {}

    num_clusters = max(entity_cluster_dict.values()) + 1

    cluster_name = {}

    for entity, cluster in entity_cluster_dict.items():
        if parent_entity_dict.get(cluster, None):
            cluster_freq = entity_frequency_dict[cluster_name[cluster]]
            parent_entity_dict[cluster].append(entity)
            if entity_frequency_dict[entity] > cluster_freq:
                cluster_name.update({cluster: entity})
        else:
            parent_entity_dict[cluster] = [entity]
            cluster_name[cluster] = entity

    for k in range(num_clusters):
        parent_cluster = cluster_name[k]
        parent_entity_dict[parent_cluster] = parent_entity_dict.pop(k)

    return parent_entity_dict


def add_entities(relation_list: list,
                 parent_entity_dict: dict,
                 filtered_messages_dict: dict,
                 node_messages_examples: int) -> list:
    """Add entities to the relation list

    Create by adding the nodes and leaves of the three. The first value of
    the sub-list is the word to be printed, the second value is the id of
    the node (leaf), the third is the id of the parent node, and last the
    color.

    The first branch is the cluster name, the second is the entity and
    the last is the sample messages.

    Parameters
    ----------
    relation_list: list
        a list with the representation root of the tree.
    parent_entity_dict: dict
        dictionary with the entity-cluster relation.
    filtered_messages_dict: dict
        dictionary with the entity-message relation.
    node_messages_examples: int
        maximum of leaves in each branch.

    Returns
    -------
    list
        a list with the representation of the tree.
    """
    ind_cluster = len(parent_entity_dict)
    ind_entities = ind_cluster + 1
    ind_message = ind_cluster + sum([len(value) for value in
                                     list(parent_entity_dict.values())]) + 1
    cluster_list = sorted_cluster(parent_entity_dict, filtered_messages_dict)
    for k in range(ind_cluster):
        cluster = cluster_list[k]
        relation_list += [[k + 1, cluster[0], 0, 1, 'white']]
        for i in range(len(cluster[1])):
            value = cluster[1][i]
            freq_others = sum([freq for _, freq in
                               cluster[2][i][node_messages_examples:]])
            freq_show = sum([freq for _, freq in
                             cluster[2][i][:node_messages_examples]])
            relation_list += [[ind_entities, value, k + 1,
                               freq_others + freq_show, 'white']]
            if freq_others < 0.6 * (freq_show + freq_others):
                for msg, freq in cluster[2][i][:node_messages_examples]:
                    relation_list += [[ind_message, msg, ind_entities,
                                       freq, 'white']]
                    ind_message += 1
                if len(cluster[2][i][node_messages_examples:]) > 0:
                    relation_list += [
                        [ind_message, 'Outras frases', ind_entities,
                         freq_others, 'white']]
                    ind_message += 1
            ind_entities += 1
    return relation_list


def sorted_cluster(parent_entity_dict: dict,
                   filtered_messages_dict: dict) -> list:
    """Sort cluster by frequency

    Parameters
    ----------
    parent_entity_dict: dict
        dictionary with the relation entity-cluster.
    filtered_messages_dict: dict
        dictionary with the relation entity-message

    Returns
    -------
    list
        a list with the clusters, entities and message sorted by frequency
    """
    output = []
    for key, value in parent_entity_dict.items():
        sorted_entity, freq, msg_list = sorted_entities(value,
                                                        filtered_messages_dict)
        aux_list = [key, sorted_entity, msg_list, freq]
        output.append(aux_list)
    return sorted(output, key=operator.itemgetter(3), reverse=True)


def sorted_entities(entities_list: list,
                    filtered_messages_dict: dict) -> tuple:
    """Sort entities by frequency


    Parameters
    ----------
    entities_list: list
        list with the entities of a cluster.
    filtered_messages_dict: list
        dictionary with the relation entity-message

    Returns
    -------
    tuple
        the sorted entities, the cumulative frequency and sorted messages.
    """
    aux_list = []
    for entity in entities_list:
        sorted_messages = sorted(filtered_messages_dict[entity],
                                 key=operator.itemgetter(1), reverse=True)

        aux_list.append([entity, sorted_messages,
                         sum(value[1] for value in sorted_messages)])
    aux_list = sorted(aux_list, key=operator.itemgetter(2), reverse=True)
    entities_sorted = [value[0] for value in aux_list]
    msg_sorted = [value[1] for value in aux_list]
    freq = sum([value[2] for value in aux_list])
    return entities_sorted, freq, msg_sorted


def create_entity_hierarchy(filtered_messages_frequency_dict: dict,
                            parent_entity_dict: dict,
                            node_messages_examples: int) -> str:
    """Create the html with the tree representation

    Parameters
    ----------
    filtered_messages_frequency_dict: dict
        dictionary with the relation entity-message
    parent_entity_dict: dict
        dictionary with the relation entity-cluster.
    node_messages_examples: int
        maximum of leaves in a branch.

    Returns
    -------
    str
        a string for the html
    """
    relation_list = [[0, 'Assuntos', -1, 1, 'white']]
    relation_list = add_entities(relation_list, parent_entity_dict,
                                 filtered_messages_frequency_dict,
                                 node_messages_examples)
    html = """<html>
      <head>
        <script type="text/javascript" src="https://www.gstatic.com/charts/loader.js"></script>
        <script type="text/javascript">
          google.charts.load('current', {packages:['wordtree']});
          google.charts.setOnLoadCallback(drawSimpleNodeChart);
          function drawSimpleNodeChart() {
            var nodeListData = new google.visualization.arrayToDataTable([
              ['id', 'childLabel', 'parent', 'size', { role: 'style' }],
              %s);

            var options = {
              backgroundColor: '#262730',
              fontName: 'Times-Roman',
              wordtree: {
                format: 'explicit',
                type: 'suffix'
              }           
            };

            var wordtree = new google.visualization.WordTree(document.getElementById('wordtree_explicit'));
            wordtree.draw(nodeListData, options);
          }
        </script>
      </head>
      <body>
        <div id="wordtree_explicit" style="width: 1500px; height: 1200px;"></div>
      </body>
    </html>""" % (str(relation_list)[1:])
    return html


def create_hierarchy_frequency_dict(entity_frequency_dict: dict,
                                    parent_entity_dict: dict) -> tp.Tuple[
    dict, dict]:
    """
    Create a dictionary as the example:

    {groups: [{name: cartão, frequency: 400, children: [{name: cartão de
    crédito, frequency: 129}, {name: cartão de débito, frequency: 81},
    name: cartão, frequency: 200}]}]}

    Parameters
    ----------
    entity_frequency_dict: dict
        dictionary with the relation entity-frequency
    parent_entity_dict: dict
        dictionary with the relation parent-entities

    Returns
    -------
    tp.Tuple[dict, dict]
        tuple with dictionary with information about the clusters and a
        dictionary with the frequency of each cluster.
    """
    parent_dict = {}
    output_list = []
    for key, value in parent_entity_dict.items():
        children_list = []
        frequency = 0
        for entity in value:
            children_list.append({'name': entity,
                                  'frequency': entity_frequency_dict[entity]})
            frequency += entity_frequency_dict[entity]

        parent_dict[key] = frequency
        hierarcy_frequency_dict = {'name': key,
                                   'frequency': frequency,
                                   'children': children_list}
        output_list.append(hierarcy_frequency_dict)
    return {'group': output_list}, parent_dict


def save_entity_hierarchy(full_path: str, entity_hierarchy_html: str) -> None:
    """
    Parameters
    ---------
    full_path : str
        String with the path to save the file
    entity_hierarchy_html : str
        HTML with the entities relation.
    """
    with open(full_path, 'w', encoding='utf-8') as f:
        f.write(entity_hierarchy_html)


def save_entity_hierarchy_dict(full_path: str,
                               entity_hierarchy_dict: dict) -> None:
    """
    Parameters
    ---------
    full_path : str
        String with the path to save the file
    entity_hierarchy_dict : dict
        dictionary with information about the clusters
    """
    with open(full_path, 'w', encoding='utf-8') as f:
        json.dump(entity_hierarchy_dict, f, ensure_ascii=False)


def save_top_plot(full_path: str, topic_freq_dict: dict) -> None:
    """
    Parameters
    ---------
    full_path : str
        String with the path to save the file
    topic_freq_dict : dict
        dictionary with the frequency of each topic.
    """
    sorted_dict = {k: v for k, v in sorted(topic_freq_dict.items(),
                                           key=lambda item: item[1],
                                           reverse=True)}
    subjects = [sub.capitalize() for sub in list(sorted_dict.keys())[:5]]
    values = list(sorted_dict.values())[:5]

    subjects.reverse()
    values.reverse()

    fig = px.bar(y=subjects,
                 x=values,
                 color_discrete_sequence=px.colors.qualitative.G10,
                 title="Analisando top 5 assuntos",
                 template='simple_white',
                 orientation='h',
                 labels={'x': 'Frequência',
                         'y': 'Assuntos'})
    fig.update_layout(autosize=False,
                      width=500,
                      height=500)
    fig.update_traces(width=0.8)
    fig.write_image(full_path)
